from crosstex.style.basic import *

# Prefer long names.
string._addproducer(makegetterproducer('shortname'), 'value')
string._addproducer(makegetterproducer('longname'), 'value')

# Preface conference names with 'Proceedings of the'.
publication._addfilter(proceedingsfilter, 'fullpublication', 'booktitle')

# Use 'In' and emphasize book and journal titles.
publication._addfilter(emphfilter, 'fullpublication', 'booktitle')
publication._addfilter(emphfilter, 'fullpublication', 'journal')
book._addfilter(emphfilter, 'fulltitle')
journal._addfilter(emphfilter, 'fulltitle')
publication._addfilter(infilter, 'fullpublication', 'booktitle')
publication._addfilter(infilter, 'fullpublication', 'journal')

# Use long labels.
publication._addproducer(makegetterproducer('fullnamelabel'), 'label')



'''
Information about the objects that can occur in databases.

The class hierarchy here describes how formatting works, which fields
are required, optional, and can derive values from other fields, etc.

REQUIRED and OPTIONAL are constants describing requirement levels for
fields in objects.
'''

import re
from copy import copy


REQUIRED = object()
OPTIONAL = object()


def cmpclasses(a, b):
    'A comparison function sorting subclasses before parents.'

    le = issubclass(a, b)
    ge = issubclass(b, a)
    if le and not ge:
        return -1
    if ge and not le:
        return 1
    return cmp(a.__name__, b.__name__)


_producers = {}
_listfilters = {}
_listformatters = {}
_filters = {}


class Formatter(object):

    _citeable = False

    def __init__(self, *kargs, **kwargs):
        self._value = None
        super(Formatter, self).__init__(*kargs, **kwargs)

    def _chain(cls, filter, chain, context):
        if not context:
            raise ValueError, 'Empty context'
        l = chain.setdefault(context, {}).setdefault(cls, [])
        v = filter
        if not hasattr(v, '__call__'):
            v = lambda obj, value, context: filter
        if chain is _producers:
            l.insert(0, v)
        else:
            l.append(v)
    _chain = classmethod(_chain)

    def _addproducer(cls, filter, *context):
        cls._chain(filter, _producers, context)
    _addproducer = classmethod(_addproducer)

    def _addlistfilter(cls, filter, *context):
        cls._chain(filter, _listfilters, context)
    _addlistfilter = classmethod(_addlistfilter)

    def _addlistformatter(cls, filter, *context):
        cls._chain(filter, _listformatters, context)
    _addlistformatter = classmethod(_addlistformatter)

    def _addfilter(cls, filter, *context):
        cls._chain(filter, _filters, context)
    _addfilter = classmethod(_addfilter)

    def _filter(self, chain, value, stop, context):
        if stop(value):
            return value
        for i in range(len(context)):
            try:
                filters = chain[context[i:]]
            except KeyError:
                continue
            kinds = [ t for t in filters if isinstance(self, t) ]
            kinds.sort(cmpclasses)
            for kind in kinds:
                for filter in filters[kind]:
                    newvalue = filter(self, value, context)
                    if newvalue is not None:
                        value = newvalue
                        if stop(value):
                            return value
        return value

    def _format(self, *context):
        listtest = lambda x: x is None or not hasattr(x, '__iter__')
        value = self._filter(_producers, None, lambda x: x is not None, context)
        value = self._filter(_listfilters, value, listtest, context)
        value = self._filter(_listformatters, value, listtest, context)
        if value is not None:
            value = str(value)
        value = self._filter(_filters, value, lambda x: x is None, context)
        if value is not None:
            value = str(value)
        return value

    def __str__(self):
        if self._value is None:
            self._value = self._format('value')
            if self._value is None:
                self._value = ''
        return self._value


class ObjectList(Formatter, list):
    def __init__(self, *kargs, **kwargs):
        super(ObjectList, self).__init__(*kargs, **kwargs)


class Object(Formatter):
    def __init__(self, keys, fields, conditions, file, line, *kargs, **kwargs):
        super(Object, self).__init__(*kargs, **kwargs)
        self.keys = keys
        self.file = file
        self.line = line
        self.kind = type(self).__name__
        self.kind = self.kind[self.kind.rfind('.')+1:]
        self.conditions = conditions
        self.fields = fields
        self.citation = None


class Concatenation(list):
    def __str__(self):
        return ''.join([str(x) for x in self])


#
# The actual objects.
#

class string(Object):
    name = ['longname', 'shortname']
    shortname = [REQUIRED, 'name', 'longname']
    longname = [REQUIRED, 'name', 'shortname']

class author(string):
    address = OPTIONAL
    affiliation = OPTIONAL
    email = OPTIONAL
    institution = OPTIONAL
    organization = OPTIONAL
    phone = OPTIONAL
    school = OPTIONAL
    url = OPTIONAL

class state(string):
    country = OPTIONAL

class country(string):
    pass

class location(string):
    name = [REQUIRED, 'longname', 'shortname', 'city', 'state', 'country']
    shortname = ['name', 'longname']
    longname = ['name', 'shortname']
    city = OPTIONAL
    state = OPTIONAL
    country = OPTIONAL

class month(string):
    monthno = REQUIRED

class journal(string):
    pass

class institution(string):
    address = OPTIONAL

class school(institution):
    pass

class publication(Object):
    _citeable = True

    abstract = OPTIONAL
    address = OPTIONAL
    affiliation = OPTIONAL
    annote = OPTIONAL
    author = OPTIONAL
    bib = OPTIONAL
    bibsource = OPTIONAL
    booktitle = OPTIONAL
    category = OPTIONAL
    subcategory = OPTIONAL
    chapter = OPTIONAL
    contents = OPTIONAL
    copyright = OPTIONAL
    crossref = OPTIONAL
    doi = OPTIONAL
    dvi = OPTIONAL
    edition = OPTIONAL
    editor = OPTIONAL
    ee = OPTIONAL
    ftp = OPTIONAL
    howpublished = OPTIONAL
    html = OPTIONAL
    http = OPTIONAL
    institution = OPTIONAL
    isbn = OPTIONAL
    issn = OPTIONAL
    journal = 'newspaper'
    newspaper = 'journal'
    key = OPTIONAL
    keywords = OPTIONAL
    language = OPTIONAL
    lccn = OPTIONAL
    location = OPTIONAL
    month = OPTIONAL
    monthno = OPTIONAL
    mrnumber = OPTIONAL
    note = OPTIONAL
    number = OPTIONAL
    organization = OPTIONAL
    pages = OPTIONAL
    pdf = OPTIONAL
    price = OPTIONAL
    ps = OPTIONAL
    publisher = OPTIONAL
    rtf = OPTIONAL
    school = OPTIONAL
    series = OPTIONAL
    size = OPTIONAL
    title = OPTIONAL
    type = OPTIONAL
    url = OPTIONAL
    volume = OPTIONAL
    year = OPTIONAL

class misc(publication):
    pass

class article(publication):
    author = REQUIRED
    title = REQUIRED
    journal = [REQUIRED, 'newspaper']
    newspaper = 'journal'
    year = REQUIRED

class book(publication):
    author = REQUIRED
    editor = OPTIONAL
    title = REQUIRED
    publisher = REQUIRED
    year = REQUIRED

class booklet(publication):
    title = REQUIRED

class inbook(publication):
    author = REQUIRED
    editor = REQUIRED
    title = REQUIRED
    chapter = REQUIRED
    pages = REQUIRED
    publisher = REQUIRED
    year = REQUIRED

class incollection(publication):
    author = REQUIRED
    title = REQUIRED
    booktitle = REQUIRED
    publisher = REQUIRED
    year = REQUIRED

class inproceedings(publication):
    author = REQUIRED
    title = REQUIRED
    booktitle = REQUIRED
    year = REQUIRED

class manual(publication):
    title = REQUIRED

class thesis(publication):
    author = REQUIRED
    title = REQUIRED
    school = REQUIRED
    year = REQUIRED

class mastersthesis(thesis):
    pass

class phdthesis(thesis):
    pass

class proceedings(publication):
    title = REQUIRED
    year = REQUIRED

class collection(proceedings):
    pass

class patent(publication):
    author = REQUIRED
    title = REQUIRED
    number = REQUIRED
    month = REQUIRED
    year = REQUIRED

class techreport(publication):
    author = REQUIRED
    title = REQUIRED
    institution = REQUIRED
    year = REQUIRED

class unpublished(publication):
    author = REQUIRED
    title = REQUIRED
    note = REQUIRED

class conference(string):
    abstract = OPTIONAL
    address = OPTIONAL
    affiliation = OPTIONAL
    annote = OPTIONAL
    author = OPTIONAL
    bib = OPTIONAL
    bibsource = OPTIONAL
    booktitle = OPTIONAL
    category = OPTIONAL
    subcategory = OPTIONAL
    chapter = OPTIONAL
    contents = OPTIONAL
    copyright = OPTIONAL
    crossref = OPTIONAL
    doi = OPTIONAL
    dvi = OPTIONAL
    edition = OPTIONAL
    editor = OPTIONAL
    ee = OPTIONAL
    ftp = OPTIONAL
    howpublished = OPTIONAL
    html = OPTIONAL
    http = OPTIONAL
    institution = OPTIONAL
    isbn = OPTIONAL
    issn = OPTIONAL
    journal = 'newspaper'
    newspaper = 'journal'
    key = OPTIONAL
    keywords = OPTIONAL
    language = OPTIONAL
    lccn = OPTIONAL
    location = OPTIONAL
    month = OPTIONAL
    monthno = OPTIONAL
    mrnumber = OPTIONAL
    note = OPTIONAL
    number = OPTIONAL
    organization = OPTIONAL
    pages = OPTIONAL
    pdf = OPTIONAL
    price = OPTIONAL
    ps = OPTIONAL
    publisher = OPTIONAL
    rtf = OPTIONAL
    school = OPTIONAL
    series = OPTIONAL
    size = OPTIONAL
    title = OPTIONAL
    type = OPTIONAL
    url = OPTIONAL
    volume = OPTIONAL
    year = OPTIONAL

class conferencetrack(conference):
    conference = OPTIONAL

class workshop(conferencetrack):
    pass

class rfc(publication):
    author = REQUIRED
    title = REQUIRED
    number = REQUIRED
    month = REQUIRED
    year = REQUIRED

class url(publication):
    url = REQUIRED
    accessmonth = OPTIONAL
    accessyear = OPTIONAL
    accessday = OPTIONAL

class newspaperarticle(article):
    author = OPTIONAL

class newspaper(journal):
    pass


#
# Initialization
#

def _fieldproducer(obj, value, context):
    try:
        return obj.fields.get(context[-1], None)
    except AttributeError:
        return
    except IndexError:
        return

for kind in globals().values():
    if not (isinstance(kind, type) and issubclass(kind, Object)):
        continue

    kind._required = {}
    kind._allowed = {}
    kind._fillin = {}

    for field in dir(kind):
        if field.startswith('_'):
            continue
        kind._allowed[field] = True
        value = getattr(kind, field)

        if hasattr(value, '__iter__'):
            if REQUIRED in value:
                kind._required[field] = True
            kind._fillin[field] = [str(f) for f in value \
                                   if f is not REQUIRED and f is not OPTIONAL]
        else:
            if value is REQUIRED:
                kind._required[field] = True
            elif value is not OPTIONAL:
                kind._fillin[field] = [str(value)]

        kind._addproducer(_fieldproducer, field)



from crosstex.style.basic import *

# Prefer long names.
string._addproducer(makegetterproducer('shortname'), 'value')
string._addproducer(makegetterproducer('longname'), 'value')

# Preface conference names with 'Proceedings of the'.
publication._addfilter(proceedingsfilter, 'fullpublication', 'booktitle')

# Use 'In' and emphasize book and journal titles.
publication._addfilter(infilter, 'fullpublication', 'booktitle')
publication._addfilter(infilter, 'fullpublication', 'journal')
publication._addfilter(emphfilter, 'fullpublication', 'booktitle')
publication._addfilter(emphfilter, 'fullpublication', 'journal')
book._addfilter(emphfilter, 'fulltitle')
journal._addfilter(emphfilter, 'fulltitle')

# Use long labels.
publication._addproducer(makegetterproducer('fullnamelabel'), 'label')

# Show extras and wrap in some HTML magic to allow popups.
publication._addproducer(extrasproducer, 'extras')

# Title first and bold.
publication._addfilter(boldfilter, 'fulltitle')
publication._addproducer(makejoinproducer(".", "\n\\newblock ", ".", "", 'fulltitle', 'fullauthorpublicationextras'), 'value')



'''
Parse files containing CrossTeX databases and citations.

Care is taken to ensure that even in the face of errors, CrossTeX does
as much as possible.  Additionally, errors are never thrown except when
necessary for the bibliography the user wants.  Entries that are not
cited or use in inheritance will never be resolved or touched beyond
the initial parsing, which is designed to be fast and error-free.
This is for efficiency as much as peace of mind.  To check every entry
in a database, simply cite every entry, and then they will all have to
be resolved and checked.
'''

import os
import re
import cPickle as pickle
import traceback
import ply.lex
import ply.yacc

from copy import copy
from crosstex.options import OptionParser, error, warning


class Value:
    'A field value containing a string, number, key, or concatenation.'

    def __init__(self, file, line, value, kind=None):
        self.file = file
        self.line = line
        if kind is not None:
            self.kind = kind
            self.value = value
        else:
            try:
                self.value = int(value)
                self.kind = 'number'
            except ValueError:
                self.value = str(value)
                self.kind = 'string'

    def concat(self, other, file, line):
        if self.kind != 'concat':
            self.value = [Value(self.file, self.line, self.value, self.kind)]
            self.kind = 'concat'
            self.file = file
            self.line = line
        if other.kind == 'concat':
            self.value.extend(other)
        else:
            self.value.append(other)

    def resolve(self, db, seen=None, trystrings=False):
        entries = []
        if self.kind == 'concat':
            for other in self.value:
                entries.extend(other.resolve(db, seen))
        elif self.kind == 'key' or (trystrings and self.kind == 'string'):
            entry = db._resolve(self.value, seen)
            if entry is not None:
                entries.append(entry)
                self.value = entry
                self.kind = 'entry'
            elif self.kind == 'key':
                error(db.options, '%s:%d: No such key "%s".' % (self.file, self.line, self.value))
                self.kind = 'string'
        return entries

    def __str__(self):
        if self.kind == 'concat':
            return ''.join([str(other) for other in self.value])
        else:
            return str(self.value)


class Entry:
    '''
    A single, raw entry in a database.

    Every piece of information is designed to avoid causing errors in the
    lexer/parser, and instead allow errors to be noticed during resolution.
    Thus, all fields (fields and defaults) are stored as lists of pairs
    rather than dictionaries, in order to preserve duplicates, etc.
    '''

    def __init__(self, kind, keys, fields, file, line, defaults={}):
        self.kind = kind
        self.keys = keys
        self.fields = fields
        self.file = file
        self.line = line
        self.defaults = defaults


class XTXFileInfo:
    'Same stuff as in Parser, but only for one file'

    def __init__(self):
        self.titlephrases = []
        self.titlesmalls = []
        self.preambles = []
        self.entries = {}
        self.tobeparsed = []

    def parse(self, file, **kwargs):
        self.tobeparsed.append(file)

    def merge(self, db):
        db.titlephrases = db.titlephrases + self.titlephrases
        db.titlesmalls = db.titlesmalls + self.titlesmalls
        db.preambles = db.preambles + self.preambles
        db.entries.update(self.entries)
        for file in self.tobeparsed:
            db.parse(file, exts=['.xtx', '.bib'])


class Parser:
    'A structure of almost raw data from the databases.'

    def __init__(self, options, optparser):
        self.optparser = optparser
        self.options = options
        self.titlephrases = []
        self.titlesmalls = []
        self.preambles = []
        self.entries = {}
        self.seen = {}
        self.cur = []

    def parse(self, file, exts=['.aux', '.xtx', '.bib']):
        'Find a file with a reasonable extension and extract its information.'

        filepath, file = os.path.split(file)
        file, fileext = os.path.splitext(file)
        if fileext in exts:
            tryexts = [fileext]
        else:
            file += fileext
            tryexts = exts

        if file in self.seen:
            if 'file' in self.options.dump:
                print 'Already processed %s.' % file
            return self.seen[file]

        if os.path.isabs(filepath):
            trydirs = [filepath]
        else:
            trydirs = [os.path.join(dir, filepath) for dir in self.options.dir]
            if self.cur:
                trydirs.append(os.path.join(self.cur[-1], filepath))
            trydirs.reverse()

        for dir in trydirs:
            for ext in tryexts:
                try:
                    self.cur.append(dir)
                    path = os.path.join(dir, file + ext)
                    try:
                        stream = open(path, 'r')
                    except IOError:
                        continue

                    if ext == '.aux':
                        if 'file' in self.options.dump:
                            print 'Processing auxiliary file %s.' % path
                        self.parseaux(stream, path)
                        return path

                    self.seen[file] = path
                    cpath = os.path.join(dir, '.' + file + ext + '.cache')
                    try:
                        if os.path.getmtime(path) < os.path.getmtime(cpath):
                            cstream = open(cpath, 'rb')
                            if 'file' in self.options.dump:
                                print "Processing database %s from cache." % path
                            try:
                                db = pickle.load(cstream)
                                cstream.close()
                                db.merge(self)
                                return path
                            except:
                                if 'file' in self.options.dump:
                                    print "Could not read cache %s, falling back to database." % cpath
                    except IOError:
                        pass
                    except OSError:
                        pass

                    if 'file' in self.options.dump:
                        print 'Processing database %s.' % path
                    db = self.parsextx(stream, path)
                    stream.close()
                    db.merge(self)

                    cpath = os.path.join(dir, '.' + file + ext + '.cache')
                    try:
                        cstream = open(cpath, 'wb')
                        try:
                            pickle.dump(db, cstream, protocol=2)
                        except:
                            if 'file' in self.options.dump:
                                print "Could not write cache %s." % cpath
                        cstream.close()
                    except IOError:
                        pass

                    return path
                finally:
                    self.cur.pop()

        error(self.options, 'Can not find %s database.' % file)
        return file

    def parseaux(self, stream, path):
        'Parse and handle options set in a TeX .aux file.'

        for line in stream:
            if line.startswith(r'\citation'):
                for citation in line[10:].strip().rstrip('}').split(','):
                    citation = citation.strip(',\t')
                    if citation:
                        self.addopts(['--cite', citation])
            elif line.startswith(r'\bibstyle'):
                self.addopts(['--style'] + line[10:].rstrip().rstrip('}').split(' '))
            elif line.startswith(r'\bibdata'):
                self.files(line[9:].rstrip().rstrip('}').split(','), exts=['.bib', '.xtx'])
            elif line.startswith(r'\@input'):
                self.files(line[8:].rstrip().rstrip('}').split(','), exts=['.aux'])

    def parsextx(self, stream, path):
        'Parse and handle options set in a CrossTeX .xtx or .bib database file.'

        db = XTXFileInfo()
        db.options = self.options
        contents = stream.read()
        if contents:
            lexer = ply.lex.lex(reflags=re.UNICODE)
            lexer.path = path
            lexer.file = os.path.basename(path)
            lexer.lineno = 1
            lexer.expectstring = False

            lexer.db = db
            lexer.defaults = {}

            parser = ply.yacc.yacc(debug=0, write_tables=0)
            parser.parse(contents, lexer=lexer)
        return db

    def files(self, files, **kwargs):
        return [ self.parse(file, **kwargs) for file in files ]

    def addopts(self, opts):
        return self.optparser.parse_args(args=opts, values=self.options)


#
# Tokens
#

newlinere = re.compile(r'\r\n|\r|\n')
numberre = re.compile(r'^\d+$')
andre = re.compile(r'\s+and\s+')

tokens = ( 'AT', 'COMMA', 'SHARP', 'OPENBRACE', 'CLOSEBRACE', 'LBRACK',
  'RBRACK', 'EQUALS', 'ATINCLUDE', 'ATSTRING', 'ATEXTEND', 'ATPREAMBLE',
  'ATCOMMENT', 'ATDEFAULT', 'ATTITLEPHRASE', 'ATTITLESMALL', 'NAME', 'NUMBER',
  'STRING', )

t_ignore = ' \t'

def t_COMMENT(t):
    r'\%.*'
    pass

def t_ATINCLUDE(t):
    r'@[iI][nN][cC][lL][uU][dD][eE]'
    t.lexer.expectstring = False
    return t

def t_ATSTRING(t):
    r'@[sS][tT][rR][iI][nN][gG]'
    t.lexer.expectstring = False
    return t

def t_ATEXTEND(t):
    r'@[eE][xX][tT][eE][nN][dD]'
    t.lexer.expectstring = False
    return t

def t_ATPREAMBLE(t):
    r'@[pP][rR][eE][aA][mM][bB][lL][eE]'
    t.lexer.expectstring = False
    return t

def t_ATCOMMENT(t):
    r'@[Cc][Oo][Mm][Mm][Ee][Nn][Tt]'
    t.lexer.expectstring = True
    return t

def t_ATDEFAULT(t):
    r'@[Dd][Ee][Ff][Aa][Uu][Ll][Tt]'
    t.lexer.expectstring = False
    return t

def t_ATTITLEPHRASE(t):
    r'@[Tt][Ii][Tt][Ll][Ee][Pp][Hh][Rr][Aa][Ss][Ee]'
    t.lexer.expectstring = True
    return t

def t_ATTITLESMALL(t):
    r'@[Tt][Ii][Tt][Ll][Ee][Ss][Mm][Aa][Ll][Ll]'
    t.lexer.expectstring = True
    return t

def t_STRING(t):
    r'"(\\.|[^\\"])*"'
    t.lexer.expectstring = False
    t.lexer.lineno += len(newlinere.findall(t.value))
    t.value = t.value[1:-1]
    return t

def t_EQUALS(t):
    r'='
    t.lexer.expectstring = True
    return t

def t_NAME(t):
    r'[-a-zA-Z:0-9/_.]+'
    t.lexer.expectstring = False
    if numberre.match(t.value):
        t.type = 'NUMBER'
    return t

def t_OPENBRACE(t):
    r'\{'
    if t.lexer.expectstring:

        bracelevel = 1
        while bracelevel > 0:
            c = t.lexer.lexdata[t.lexer.lexpos]
            t.lexer.lexpos += 1
            if c == '{' and t.value[-1] != '\\':
                bracelevel += 1
            if c == '}' and t.value[-1] != '\\':
                bracelevel -= 1
            t.value += c

        t.lexer.expectstring = False
        t.lexer.lineno += len(newlinere.findall(t.value))
        t.value = t.value[1:-1]
        if numberre.match(t.value):
            t.type = 'NUMBER'
        else:
            t.type = 'STRING'

    return t

def t_CLOSEBRACE(t):
    r'\}'
    t.lexer.expectstring = False
    return t

def t_AT(t):
    r'@'
    t.lexer.expectstring = False
    return t

def t_SHARP(t):
    r'\#'
    t.lexer.expectstring = True
    return t

def t_COMMA(t):
    r','
    t.lexer.expectstring = False
    return t

def t_LBRACK(t):
    r'\['
    t.lexer.expectstring = False
    return t

def t_RBRACK(t):
    r'\]'
    t.lexer.expectstring = False
    return t

def t_newline(t):
    r'(\r\n|\r|\n)'
    t.lexer.lineno += 1

def t_error(t):
    error(t.lexer.db.options, '%s:%d: Syntax error near "%s".' %
          (t.lexer.file, t.lexer.lineno, t.value[:20]))
    t.skip(1)


#
# Grammar; start symbol is stmts.
#

precedence = ( )

def p_ignore(t):
    '''
    stmts :
          | stmt stmts
    stmt  : ATCOMMENT STRING
    '''

def p_stmt_preamble(t):
    'stmt : ATPREAMBLE OPENBRACE STRING CLOSEBRACE'
    t.lexer.db.preambles.append(t[3])

def p_stmt_titlephrase(t):
    'stmt : ATTITLEPHRASE STRING'
    t.lexer.db.titlephrases.append(t[2])

def p_stmt_titlesmall(t):
    'stmt : ATTITLESMALL STRING'
    t.lexer.db.titlesmalls.append(t[2])

def p_stmt_include(t):
    'stmt : ATINCLUDE NAME'
    t.lexer.db.parse(t[2], exts=['.xtx', '.bib'])

def p_stmt_default(t):
    'stmt : ATDEFAULT field'
    t.lexer.defaults = copy(t.lexer.defaults)
    if t[2][1]:
        t.lexer.defaults[t[2][0]] = t[2][1]
    else:
        try:
            del t.lexer.defaults[t[2][0]]
        except KeyError:
            pass

def p_stmt_string(t):
    'stmt : ATSTRING OPENBRACE fields CLOSEBRACE'
    file, line, defaults = t.lexer.file, t.lineno(2), t.lexer.defaults
    for key, value in t[3]:
        ent = Entry('string', [key], [([], [('name', value)])],
                    file, line, defaults)
        t.lexer.db.entries.setdefault(key, []).append(ent)

def p_stmt_entry(t):
    'stmt : entry'
    for key in t[1].keys:
        t.lexer.db.entries.setdefault(key, []).append(t[1])

def p_entry(t):
    'entry : AT NAME OPENBRACE keys COMMA conditionals CLOSEBRACE'
    file, line, defaults = t.lexer.file, t.lineno(2), t.lexer.defaults
    t[0] = Entry(t[2].lower(), t[4], t[6], file, line, defaults)

def p_entry_extend(t):
    '''
    entry : ATEXTEND OPENBRACE keys COMMA conditionals CLOSEBRACE
          | ATEXTEND OPENBRACE keys CLOSEBRACE
    '''
    try:
        fields = t[5]
    except IndexError:
        fields = []
    file, line, defaults = t.lexer.file, t.lineno(1), t.lexer.defaults
    t[0] = Entry('extend', t[3], fields, file, line, defaults)

def p_keys_singleton(t):
    'keys : NAME'
    t[0] = [t[1]]

def p_keys_multiple(t):
    'keys : NAME EQUALS keys'
    t[0] = [t[1]] + t[3]

def p_conditionals_empty(t):
    'conditionals :'
    t[0] = []

def p_conditionals_singleton(t):
    'conditionals : conditional'
    t[0] = [t[1]]

def p_conditionals_multiple(t):
    'conditionals : conditional conditionals'
    t[0] = [t[1]] + t[2]

def p_conditionals_unconditional(t):
    'conditionals : fields conditionals'
    t[0] = [([], t[1])] + t[2]

def p_conditional(t):
    'conditional : LBRACK fields RBRACK fields'
    t[0] = (t[2], t[4])

def p_fields_empty(t):
    'fields :'
    t[0] = []

def p_fields_singleton(t):
    'fields : field'
    t[0] = [t[1]]

def p_fields(t):
    'fields : field COMMA fields'
    t[0] = [t[1]] + t[3]

def p_field(t):
    'field : NAME EQUALS value'
    t[0] = (t[1].lower(), t[3])

def p_value_singleton(t):
    'value : simplevalue'
    t[0] = t[1]

def p_value_concat(t):
    'value : value SHARP simplevalue'
    warning(t.lexer.db.options, '%s:%d: The implementation of concatenation is currently broken.' %
          (t.lexer.file, t.lexer.lineno, t.value))
    t[0] = t[1]
    t[0].concat(t[3])

def p_simplevalue_name(t):
    'simplevalue : NAME'
    t[0] = Value(t.lexer.file, t.lineno(1), t[1], 'key')

def p_simplevalue_number(t):
    'simplevalue : NUMBER'
    t[0] = Value(t.lexer.file, t.lineno(1), t[1])

def p_simplevalue_string(t):
    'simplevalue : STRING'
    t[0] = Value(t.lexer.file, t.lineno(1), t[1])

def p_error(t):
    ply.yacc.errok()
    error(t.lexer.db.options, '%s:%d: Parse error near "%s".' %
          (t.lexer.file, t.lexer.lineno, t.value[:20]))



from crosstex.style.common import *

# Standard producers for top-level objects.
publication._addproducer(crosstexproducer, 'value')
string._addproducer(crosstexproducer, 'value')
location._addproducer(crosstexproducer, 'value')
ObjectList._addlistformatter(andcrosstexlistformatter, 'value')

# Use empty labels (irrelevant, but labels determine what is 'citeable' and thus produced.
publication._addproducer(emptyproducer, 'label')
string._addproducer(emptyproducer, 'label')
location._addproducer(emptyproducer, 'label')



from crosstex.style.basic import *

# Prefer long names.
string._addproducer(makegetterproducer('shortname'), 'value')
string._addproducer(makegetterproducer('longname'), 'value')

# Use numeric labels.
publication._addproducer(emptyproducer, 'label')

# Italicize book and journal titles.
publication._addfilter(emphfilter, 'fullpublication', 'booktitle')
publication._addfilter(emphfilter, 'fullpublication', 'journal')
book._addfilter(emphfilter, 'fulltitle')
journal._addfilter(emphfilter, 'fulltitle')

# Use alphabetic labels.
publication._addproducer(makegetterproducer('initialslabel'), 'label')



'''
Parse files containing CrossTeX databases and citations.

Care is taken to ensure that even in the face of errors, CrossTeX does
as much as possible.  Additionally, errors are never thrown except when
necessary for the bibliography the user wants.  Entries that are not
cited or use in inheritance will never be resolved or touched beyond
the initial parsing, which is designed to be fast and error-free.
This is for efficiency as much as peace of mind.  To check every entry
in a database, simply cite every entry, and then they will all have to
be resolved and checked.
'''

import re
import math
import crosstex.objects

from copy import copy
from crosstex.options import warning, error


class Constraint(dict):
    'A convenience representation of a constrained citation.'

    def __init__(self, key):
        parts = key[1:].lower().split(':')
        for i in range(len(parts)):
            constraint = parts[i].split('=', 1)
            if len(constraint) == 1:
                try:
                    values = int(constraint[0])
                except ValueError:
                    if i == 0:
                        field = 'author'
                    else:
                        field = 'title'
                    values = constraint[0].split('-')
                else:
                    field = 'year'
                    values = [str(values)]
            else:
                field = constraint[0]
                values = constraint[1].split('-')
            if not field:
                error(self.options, 'Empty field in constraint "%s".' % key)
            if not values:
                error(self.options, 'Empty values in constraint "%s".' % key)
            if field and values:
                self[field] = values

    def match(self, entry):
        for field, values in self.iteritems():
            try:
                actual = str(entry.fields[field]).lower()
            except KeyError:
                return False
            for value in values:
                if value not in actual:
                    return False
        return True


class Database(list):
    'A structure of almost raw data from the databases.'

    def __init__(self, parser):
        self.titlephrases = parser.titlephrases
        self.titlesmalls = parser.titlesmalls
        self.preambles = parser.preambles
        self.entries = parser.entries
        self.options = parser.options
        self.resolved = {}

        semantic = []
        all = not self.options.cite
        for key in self.options.cite:
            if key == '*':
                all = True
            elif key.startswith('!'):
                semantic.append(key)
            else:
                entry = self._resolve(key)
                if entry is None:
                    continue
                if not entry._citeable:
                    error(self.options, '%s is not a citeable object.' % key)
                elif entry.citation is None:
                    entry.citation = key
                    self.append(entry)
                elif entry.citation != key:
                    error(self.options,
                          'Citations "%s" and "%s" refer to the same object.' %
                          (entry.citation, key))

        entries = {}
        if all or semantic:
            for key in self.entries:
                entry = self._resolve(key)
                if entry is not None and entry._citeable:
                    entries[entry] = True

        for key in semantic:
            constraint = Constraint(key)
            if not constraint:
                error(self.options, 'Empty constraint.')
                continue
            matches = [entry for entry in entries if constraint.match(entry)]
            if not matches:
                error(self.options, 'Constraint "%s" has no matches.' % key)
                continue
            if matches[0].citation is None:
                matches[0].citation = key
                self.append(matches[0])
            elif matches[0].citation != key:
                error(self.options,
                      'Citations "%s" and "%s" refer to the same object.' %
                      (matches[0].citation, key))
            if len(matches) > 1:
                error(self.options, 'Ambiguous constraint "%s" has %d matches:' %
                      (key, len(matches)))
                error(self.options, '  accepted key %s.' % (matches[0].keys[0]))
                for match in matches[1:]:
                    error(self.options, '  rejected key %s.' % (match.keys[0]))

        if all:
            for entry in entries:
                if entry.citation is None:
                    entry.citation = entry.keys[0]
                    self.append(entry)

        self._sort()

    def dump(self, kind):
        entries = {}
        kind = getattr(crosstex.objects, kind)
        for key in self.entries:
            entry = self._resolve(key)
            if entry is not None and isinstance(entry, kind):
                entries[str(entry)] = True
        return entries.keys()

    def write(self, out, entries=None, headings=None):
        'Print the formatted database to the stream out.'

        if entries is None:
            entries = self
        if headings is None:
            headings = self.options.heading
        if (self.options.convert == 'html' or self.options.convert == 'bbl') and len(headings) > 1:
            error(self.options, 'The output format does not support multiple heading levels.')
            headings = headings[:1]

        if not headings:
            if self.options.convert == 'html' or self.options.convert == 'bbl':
                max = ''
                for entry in entries:
                    label = entry._format('label')
                    label = label.replace(r'{\etalchar{+}}', 'X')
                    if len(label) > len(max):
                        max = label
                if not max:
                    max = '0' * int(math.ceil(math.log(len(entries) + 1, 10)))
                out.write(r'\begin{thebibliography}{' + max + '}\n')
            for entry in entries:
                out.write(str(entry))
            if self.options.convert == 'html' or self.options.convert == 'bbl':
                out.write(r'\end{thebibliography}' + '\n')
            return

        heading = headings[0][1]
        cats = {}
        for entry in entries:
            cats.setdefault(str(entry.fields.get(heading, '')), []).append(entry)

        values = [k for k in cats.keys() if k != '']
        values.sort()
        if headings[0][0]:
            values.reverse()
        if '' in cats:
            values.insert(0, '')

        for value in values:
            if (self.options.convert == 'html' or self.options.convert == 'bbl') and value != '':
                out.write(r'{\renewcommand{\refname}{' + str(value) + '}\n')
            elif self.options.convert == 'xtx':
                default = crosstex.style.common._fieldval(heading, value)
                out.write('@default ' + default + ' \n\n')
            self.write(out, cats[value], headings[1:])
            if (self.options.convert == 'html' or self.options.convert == 'bbl') and value != '':
                out.write('}\n')

    def _sort(self):
        citations = self
        for reverse, field in self.options.sort:
            def keyfunc(x):
                value = x._format('sort', field)
                if value is None:
                    return x._format('sort', 'key')
                try:
                    return int(value)
                except ValueError:
                    return value
            if reverse:
                sortlist = [ (keyfunc(x), len(citations) - i, x) for i, x in enumerate(citations) ]
                sortlist.sort()
                sortlist.reverse()
            else:
                sortlist = [ (keyfunc(x), i, x) for i, x in enumerate(citations) ]
                sortlist.sort()
            citations = [ x[-1] for x in sortlist ]
        self[:] = citations

    def _get(self, key):
        '''
        Resolve a key to a single entry and all its aliases and extensions.

        This involves looking up each entry and extension of that key,
        following out through aliases.  Complain if this process reveals
        two objects aliased to each other (note that this can happen through
        multiple levels of extensions, and need not be immediately obvious).
        '''

        keys = {}
        seen = {}
        work = {key: True}
        base = None
        extensions = []
        while work:
            key, value = work.popitem()
            keys[key] = True
            for entry in self.entries.get(key, []):
                if entry in seen:
                    continue
                seen[entry] = True
                if entry.kind == 'extend':
                    extensions.append(entry)
                elif base is None:
                    base = entry
                else:
                    error(self.options, '%s:%d: Alias %s is also defined at %s:%d.' %
                          (base.file, base.line, key, entry.file, entry.line))
                    continue
                newwork = [k for k in entry.keys if k not in keys]
                work.update(dict.fromkeys(newwork, True))

        if base is None:
            if extensions:
                error(self.options, '%s is extended but never defined.' % key)
            else:
                error(self.options, '%s is never defined.' % key)
        return (keys.keys(), base, extensions)

    def _resolve(self, key, seen=None):
        '''
        Resolve the entry referenced by the given key.

        Evaluate conditional fields, inheritance, @extend entries, etc. until
        the entry is stable, and return the result.  The resolved entry is
        also memoized, so it is efficient to call this twice on one key or
        aliased keys.
        '''

        # Return memoized results.
        try:
            return self.resolved[key]
        except KeyError:
            pass

        # Find all entries relevant to the given key while guarding against
        # inheritance loops.
        keys, base, extensions = self._get(key)
        if base is None or not keys:
            return None
        if seen is None:
            seen = {}
        elif base in seen:
            error(self.options, '%s inherits itself recursively.' % key)
            return None
        seen[base] = True

        # Get the class corresponding to this object type.
        try:
            kind = getattr(crosstex.objects, base.kind)
        except AttributeError:
            error(self.options, '%s:%d: There is no such thing as a @%s.\n' %
                  (base.file, base.line, base.kind))
            return None
        if not (isinstance(kind, type) and issubclass(kind, crosstex.objects.Object)):
            error(self.options, '%s:%d: There is no such thing as a @%s.\n' %
                  (base.file, base.line, base.kind))
            return None

        # Resolve fields.  Put in conditions a list of lists of conditionals,
        # each from an entry.  Now is when to complain about duplicated
        # fields. Note that the conditions for each entry are flatly
        # concatenated into the whole list, in order of preference (extensions
        # first, then the base entry, then finally inherited objects in the
        # order in which they are assigned).
        conditions = []
        for entry in extensions + [base]:
            for condition, assignments in entry.fields:
                fieldsdict = {}
                for field, value in assignments:
                    field = field.lower()
                    if field in fieldsdict:
                        warning(self.options, '%s:%d: %s field is duplicated.' %
                                (value.file, value.line, field))
                    else:
                        fieldsdict[field] = value

                # Inherit fields, e.g. longname from name.
                for field in kind._fillin:
                    if field in fieldsdict:
                        continue
                    for fill in kind._fillin[field]:
                        if fill in fieldsdict:
                            fieldsdict[field] = fieldsdict[fill]
                            break

                conditiondict = {}
                for field, value in condition:
                    field = field.lower()
                    if field in conditiondict:
                        warning(self.options, '%s:%d: Conditional %s field is duplicated.' %
                                (value.file, value.line, field))
                    else:
                        value.resolve(self)
                        value = str(value)
                        conditiondict[field] = value
                conditions.append((conditiondict, fieldsdict, entry))

        # Do the resolution, adding inherited entries as they are assigned.
        # When the fields stabilize, add in the defaults as a last resort and
        # continue until the fields are re-stabilized.
        defaulted = False
        inherited = {}
        fields = {}
        i = 0
        while i < len(conditions) or not defaulted:
            if i >= len(conditions):
                conditions.append(({}, base.defaults, copy(base)))
                defaulted = True
                i = 0

            condition, assignments, source = conditions[i]
            matches = False
            for field, value in condition.iteritems():
                try:
                    if str(fields[field]) != value:
                        del conditions[i]
                        break
                except KeyError:
                    i += 1
                    break
            else:
                matches = True
                del conditions[i]
                i = 0
            if not matches:
                continue

            for field, value in assignments.iteritems():
                # Do not assign to fields that do not exist.
                # Complain only if set directly in base or extensions.
                if field not in kind._allowed:
                    if source is base or source in extensions:
                        warning(self.options, '%s:%d: No such field %s in %s.' %
                                (base.file, base.line, field, base.kind))
                    continue

                # Do not assign fields twice.
                if field in fields:
                    continue

                for entry in value.resolve(self, seen):
                    if entry not in inherited:
                        inherited[entry] = True
                        conditions.append(({}, entry.fields, entry))
                        conditions.extend(entry.conditions)

                if field in ['author', 'editor']:
                    authors = crosstex.objects.ObjectList()
                    oldcheck = self.options.check
                    self.options.check = 5 # Ignore everything
                    for part in re.split('\s+and\s+', str(value)): # XXX split on word boundaries properly
                        entry = self._resolve(part, seen)
                        if entry is None:
                            authors.append(part)
                        else:
                            if entry not in inherited:
                                inherited[entry] = True
                                conditions.append(({}, entry.fields, entry))
                                conditions.extend(entry.conditions)
                            authors.append(entry)
                    self.options.check = oldcheck
                    fields[field] = authors
                else:
                    fields[field] = value

        # Ensure required fields are all provided.
        for field in kind._required:
            if field not in fields:
                error(self.options, '%s:%d: Missing required field %s in %s.' %
                      (base.file, base.line, field, base.kind))

        # Memoize the result and return it.
        ent = kind(keys, fields, conditions, base.file, base.line)
        for key in keys:
            self.resolved[key] = ent

        return ent



from crosstex.objects import *

_wordre = re.compile('(-+|\s+|\\\w+|\\\W+|\$[^\$]*\$|[{}])', re.IGNORECASE)
_spacere = re.compile(r'^\s*$')
_specialre = re.compile(r'^(\\.*|\$[^\$]*\$)$')
_punctuationre = re.compile('([:!.?]|-{2,})[\'}\n]*$')
_linkre = re.compile("[a-zA-Z][-+.a-zA-Z0-9]*://([:/?#[\]@!$&'()*+,;=a-zA-Z0-9_\-.~]|%[0-9a-fA-F][0-9a-fA-F]|\\-|\s)*")
_linksub = re.compile('\\-\s')
_protectre = re.compile(r'[\\{}]')
_endre = re.compile(r"(\\end\{[^}]*\}|['\s}])*$")

_bibtexkinds = ['article', 'book', 'booklet', 'conference', 'inbook', \
  'incollection', 'inproceedings', 'manual', 'mastersthesis', 'phdthesis', \
  'proceedings', 'techreport', 'unpublished']

# Clean up a string to be a search for the literal
def _sanitize(r):
    value = ''
    for char in r:
        if char == '^':
            value += r'\^'
        else:
            value += '[' + char + ']'
    return value

# Piece an entry together.
def _punctuate(string, punctuation='', tail=' '):
    if string == None:
        string = ''
    i = _endre.search(string).start()
    end = string[i:]
    string = string[:i]
    if string and not (string.endswith('?') or string.endswith('!') or string.endswith(':') or string.endswith('--') or string.endswith(punctuation)):
        string += punctuation
    if string or end:
        end += tail
    return string + end

def _names(name, short=False, plain=False):
    value = ''
    lastchar = ' '
    names = []
    nesting = 0
    if isinstance(name, Formatter):
        name = name._format('value') or ''
    for i in range(0, len(name)):
        charc = name[i]
        if nesting == 0 and lastchar != '\\' and lastchar != ' ' and charc == ' ':
            names.append(value)
            value = ''
        elif lastchar != '\\' and charc == '}':
            if not plain:
                value += charc
            if nesting == 0:
                names.append(value)
                value = ''
            else:
                nesting -= 1
        elif lastchar != '\\' and charc == '{':
            if not plain:
                value += charc
            nesting += 1
        elif nesting == 0 and lastchar != '\\' and charc == ',':
            pass
        else:
            if not plain or (charc != '\\' and lastchar != '\\'):
                value += charc
        lastchar = charc
    names.append(value)

    # extract lastname, check suffixes and last name modifiers
    # extract also a list of first names
    snames = ['Jr.', 'Sr.', 'Jr', 'Sr', 'III', 'IV']
    mnames = ['van', 'von', 'de', 'bin', 'ibn']
    sname = ''
    snameoffset = len(names)
    while snameoffset > 0 and names[snameoffset - 1] in snames:
        snameoffset -= 1
    mnameoffset = 0
    while mnameoffset < snameoffset and names[mnameoffset] not in mnames:
        mnameoffset += 1
    lnameoffset = snameoffset
    while lnameoffset > 0 and names[lnameoffset - 1] not in mnames:
        lnameoffset -= 1
    if lnameoffset <= mnameoffset:
        lnameoffset = mnameoffset = snameoffset - 1

    # return the person info as a tuple
    (fnames, mnames, lnames, snames) = (names[:mnameoffset], names[mnameoffset:lnameoffset], names[lnameoffset:snameoffset], names[snameoffset:])
    if short:
        fnamesabbr = []
        for n in fnames:
            abbr = ''
            initial = 0
            sep = ''
            while initial < len(n):
                if n[initial] == '\\':
                    initial += 1
                elif n[initial] in '{}':
                    pass
                elif n[initial] == '~':
                    abbr += n[initial]
                elif n[initial] in '-.':
                    sep = n[initial]
                elif sep != None:
                    if sep != '.':
                        abbr += sep
                    abbr += n[initial] + '.'
                    sep = None
                initial += 1
            if abbr:
                fnamesabbr.append(abbr)
        return (['~'.join(fnamesabbr)], mnames, lnames, snames)
    else:
        return (fnames, mnames, lnames, snames)

def _last_initials(name, size):
    (fnames, mnames, lnames, snames) = _names(name)
    mnamestr = ''
    for mname in mnames:
        first = 0
        while first < len(mname):
            if mname[first] not in '{}\\':
                mnamestr += mname[first]
                break
            elif mname[first] == '\\':
                first += 2
            else:
                first += 1
    lnamestr = ''
    for lname in lnames:
        if len(lnamestr) >= size:
            break
        first = 0
        while first < len(lname):
            if lname[first] not in '{}\\':
                lnamestr += lname[first]
                if mnamestr != '' or len(lnamestr) >= size:
                    break
                else:
                    first += 1
            elif lname[first] == '\\':
                first += 2
            else:
                first += 1
    return mnamestr + lnamestr

def _fieldval(field, value):
    if isinstance(value, Object):
        return '%s = %s' % (field, value.citation)
    else:
        value = str(value)
        try:
            return '%s = %d' % (field, int(value))
        except:
            if '{' in value or '}' in value:
                return '%s = {%s}' % (field, value)
            else:
                return '%s = "%s"' % (field, value)

def makegetterproducer(field):
    def getterproducer(obj, value, context):
        return obj._format(*(context + (field,)))
    return getterproducer

def makejoinproducer(punctuation, space, final, finalspace, *fields):
    def joinproducer(obj, value, context):
        value = ''
        for field in fields:
            fieldvalue = obj._format(*(context + (field,)))
            if fieldvalue:
                value = _punctuate(value, punctuation, space) + fieldvalue
        return _punctuate(value, final, finalspace)
    return joinproducer

def bibtexproducer(obj, value, context):
    kind = obj.kind
    if kind not in _bibtexkinds:
        kind = 'misc'
    value = '@%s{%s' % (kind, obj.citation)
    for field in obj.fields:
        fieldvalue = obj._format(*(context + (field,)))
        if fieldvalue:
            value += ',\n\t' + _fieldval(field, fieldvalue)
    value += '}\n\n'
    return value

def crosstexproducer(obj, value, context):
    if not obj.keys:
        return ''
    value = '@%s{%s' % (obj.kind, ' = '.join(obj.keys))
    for field in obj.fields:
        value += ',\n\t' + _fieldval(field, obj.fields[field])
    value += '}\n\n'
    return value

def makelinksproducer(fields):
    def linksproducer(obj, value, context):
        links = ''
        for field in fields:
            myfield = field.lower()
            fieldvalue = obj._format(*(context + (myfield,)))
            if fieldvalue:
                for m in _linkre.finditer(str(fieldvalue)):
                    uri = m.group()
                    _linksub.sub(uri, '')
                    links = _punctuate(links) + '\\href{%s}{\\small\\textsc{%s}}' % (uri, field)
        return links
    return linksproducer

def extrasproducer(obj, value, context):
    extras = ''
    abstractvalue = obj._format(*(context + ('abstract',)))
    keywordsvalue = obj._format(*(context + ('keywords',)))
    if abstractvalue:
        extras = _punctuate(extras, '\n', tail='') + '\\noindent\\begin{small}%s\\end{small}' % abstractvalue
    if keywordsvalue:
        extras = _punctuate(extras, '\n\n', tail='') + '\\noindent\\begin{small}\\textsc{Keywords:} %s\\end{small}' % keywordsvalue
    if extras:
        extras = '\\begin{quotation}' + extras + '\\end{quotation}'
    return extras

def authoryearproducer(obj, value, context):
    authorvalue = obj._format(*(context + ('author',)))
    yearvalue = obj._format(*(context + ('year',)))
    if yearvalue == None:
        yearvalue = ''
    else:
        yearvalue = str(yearvalue)
    if authorvalue == None:
        authorvalue = ''
    else:
        authorvalue = str(authorvalue)
    return authorvalue + yearvalue or None

def longauthoryearproducer(obj, value, context):
    authorvalue = obj._format(*(context + ('author',)))
    yearvalue = obj._format(*(context + ('year',)))
    if yearvalue == None:
        yearvalue = ''
    else:
        yearvalue = str(yearvalue)
    if authorvalue == None:
        authorvalue = ''
    else:
        authorvalue = str(authorvalue)
    return _punctuate(authorvalue) + yearvalue or None

def authoreditorproducer(obj, value, context):
    authorvalue = obj._format(*(context + ('author',)))
    if authorvalue == None:
        authorvalue = obj._format(*(context + ('editor',)))
        if authorvalue == None:
            return None
        if ' and ' in authorvalue:
            authorvalue = str(authorvalue) + ', eds.'
        else:
            authorvalue = str(authorvalue) + ', ed.'
    else:
        authorvalue = str(authorvalue)
    return authorvalue

def dateproducer(obj, value, context):
    value = ''
    monthvalue = obj._format(*(context + ('month',)))
    if monthvalue:
        value = _punctuate(value, ',') + str(monthvalue)
    yearvalue = obj._format(*(context + ('year',)))
    if yearvalue:
        value = _punctuate(value, ',') + str(yearvalue)
    return value

def fullpublicationproducer(obj, value, context):
    value = obj._format(*(context + ('publication',)))
    booktitlevalue = obj._format(*(context + ('booktitle',)))
    journalvalue = obj._format(*(context + ('journal',)))
    if booktitlevalue:
        value = _punctuate(value, '.') + str(booktitlevalue)
        volumevalue = obj._format(*(context + ('volume',)))
        if volumevalue:
            value += ', volume %s' % volumevalue
            seriesvalue = obj._format(*(context + ('series',)))
            if seriesvalue :
                value += ' of \\emph{%s}' % seriesvalue
        chaptervalue = obj._format(*(context + ('chapter',)))
        if chaptervalue:
            value += ', chapter %s' % chaptervalue
    elif journalvalue:
        value = _punctuate(value, ',') + str(journalvalue)
        numbervalue = obj._format(*(context + ('number',)))
        volumevalue = obj._format(*(context + ('volume',)))
        pagesvalue = obj._format(*(context + ('pages',)))
        if numbervalue or volumevalue or pagesvalue:
            value = _punctuate(value, ',')
        if volumevalue:
            value += str(volumevalue)
        if numbervalue:
            value += '(%s)' % numbervalue
        if pagesvalue:
            if volumevalue or numbervalue:
                value += ':%s' % pagesvalue
            elif pagesvalue :
                value += 'page %s' % pagesvalue
            else:
                try:
                    pagenum = int(pagesvalue)
                    value += 'page %d' % pagenum
                except ValueError:
                    value += 'pages %s' % pagesvalue
    institutionvalue = obj._format(*(context + ('institution',)))
    if institutionvalue:
        value = _punctuate(value, ',') + str(institutionvalue)
    schoolvalue = obj._format(*(context + ('school',)))
    if schoolvalue:
        value = _punctuate(value, ',') + str(schoolvalue)
    if not journalvalue:
        numbervalue = obj._format(*(context + ('number',)))
        if numbervalue:
            value = _punctuate(value, ',')
            numbertypevalue = obj._format(*(context + ('numbertype',)))
            if numbertypevalue:
                value = _punctuate(value + str(numbertypevalue))
            value += str(numbervalue)
        pagesvalue = obj._format(*(context + ('pages',)))
        if pagesvalue:
            try:
                pagenum = int(pagesvalue)
                value = _punctuate(value, ',') + ('page %d' % pagenum)
            except ValueError:
                value = _punctuate(value, ',') + ('pages %s' % pagesvalue)
    authorvalue = obj._format(*(context + ('author',)))
    editorvalue = obj._format(*(context + ('editor',)))
    if authorvalue and editorvalue:
        value = _punctuate(value, ',') + str(editorvalue)
    publishervalue = obj._format(*(context + ('publisher',)))
    if publishervalue:
        value = _punctuate(value, ',') + str(publishervalue)
    addressvalue = obj._format(*(context + ('address',)))
    if addressvalue:
        value = _punctuate(value, ',') + str(addressvalue)
    notevalue = obj._format(*(context + ('note',)))
    if notevalue:
        value = _punctuate(value, ',') + str(notevalue)
    yearvalue = obj._format(*(context + ('year',)))
    if yearvalue:
        value = _punctuate(value, ',')
        monthvalue = obj._format(*(context + ('month',)))
        if monthvalue:
            value = _punctuate(value + str(monthvalue))
        value += str(yearvalue)
    return value

def acmfullpublicationproducer(obj, value, context):
    value = obj._format(*(context + ('publication',)))
    booktitlevalue = obj._format(*(context + ('booktitle',)))
    journalvalue = obj._format(*(context + ('journal',)))
    if booktitlevalue:
        value = _punctuate(value, '.') + str(booktitlevalue)
        volumevalue = obj._format(*(context + ('volume',)))
        if volumevalue:
            value += ', vol. %s' % volumevalue
            seriesvalue = obj._format(*(context + ('series',)))
            if seriesvalue :
                value += ' of \\emph{%s}' % seriesvalue
        chaptervalue = obj._format(*(context + ('chapter',)))
        if chaptervalue:
            value += ', chap. %s' % chaptervalue
    elif journalvalue:
        value = _punctuate(value, ',') + str(journalvalue)
        volumevalue = obj._format(*(context + ('volume',)))
        if volumevalue:
            value = _punctuate(value) + '\\emph{%s}' % str(volumevalue)
        numbervalue = obj._format(*(context + ('number',)))
        if numbervalue:
            value = _punctuate(value, ',') + str(numbervalue)

    addryearvalue = ''
    addressvalue = obj._format(*(context + ('address',)))
    if addressvalue:
        addryearvalue = _punctuate(addryearvalue, ',') + str(addressvalue)
    datevalue = ''
    monthvalue = obj._format(*(context + ('month',)))
    if monthvalue:
        datevalue = _punctuate(datevalue) + str(monthvalue)
    yearvalue = obj._format(*(context + ('year',)))
    if yearvalue:
        datevalue = _punctuate(datevalue) + str(yearvalue)
    if datevalue:
        addryearvalue = _punctuate(addryearvalue, ',') + str(datevalue)
    if addryearvalue:
        value = _punctuate(value) + '(%s)' % addryearvalue

    institutionvalue = obj._format(*(context + ('institution',)))
    if institutionvalue:
        value = _punctuate(value, ',') + str(institutionvalue)
    schoolvalue = obj._format(*(context + ('school',)))
    if schoolvalue:
        value = _punctuate(value, ',') + str(schoolvalue)

    publishervalue = obj._format(*(context + ('publisher',)))
    if publishervalue:
        value = _punctuate(value, ',') + str(publishervalue)

    authorvalue = obj._format(*(context + ('author',)))
    editorvalue = obj._format(*(context + ('editor',)))
    if authorvalue and editorvalue:
        value = _punctuate(value, ',') + str(editorvalue)

    pagesvalue = obj._format(*(context + ('pages',)))
    if pagesvalue:
        if not journalvalue:
            try:
                pagenum = int(pagesvalue)
                pagesvalue = 'p. ' + pagenum
            except ValueError:
                pagesvalue = 'pp. ' + pagesvalue
        value = _punctuate(value, ',') + str(pagesvalue)
    notevalue = obj._format(*(context + ('note',)))
    if notevalue:
        value = _punctuate(value, ',') + str(notevalue)
    return value

def accessedproducer(obj, value, context):
    urlvalue = str(obj._format(*(context + ('url',))))
    dayvalue = obj._format(*(context + ('accessday',)))
    yearvalue = obj._format(*(context + ('accessyear',)))
    monthvalue = obj._format(*(context + ('accessmonth',)))
    if yearvalue or monthvalue:
        urlvalue = _punctuate(urlvalue, ',') + 'Accessed'
    if monthvalue:
        urlvalue = _punctuate(urlvalue) + monthvalue
        if dayvalue:
            urlvalue = _punctuate(urlvalue) + dayvalue
        if yearvalue:
            urlvalue = _punctuate(urlvalue, ',') + yearvalue
    elif yearvalue:
        urlvalue = _punctuate(urlvalue) + yearvalue
    return urlvalue

def citystatecountryproducer(obj, value, context):
    cityvalue = obj._format(*(context + ('city',)))
    statevalue = obj._format(*(context + ('state',)))
    countryvalue = obj._format(*(context + ('country',)))
    value = ''
    if cityvalue:
        value = _punctuate(value, ',') + str(cityvalue)
    if statevalue:
        value = _punctuate(value, ',') + str(statevalue)
    if countryvalue:
        value = _punctuate(value, ',') + str(countryvalue)
    return value

def thesistypeproducer(obj, value, context):
    typevalue = obj._format(*(context + ('type',)))
    if typevalue:
        return str(typevalue)
    typevalue = obj._format(*(context + ('thesistype',)))
    if typevalue:
        return _punctuate(typevalue) + 'Thesis'
    return None

def emptyproducer(obj, value, context):
    return ''

def lastfirstfilter(obj, objvalue, context):
    (fnames, mnames, lnames, snames) = _names(objvalue)
    namestr = ''
    for n in mnames:
        namestr = _punctuate(namestr) + n
    for n in lnames:
        namestr = _punctuate(namestr) + n
    if len(fnames) > 0:
        namestr = _punctuate(namestr, ',')
    for n in fnames:
        namestr = _punctuate(namestr) + n
    for n in snames:
        namestr = _punctuate(namestr) + n
    return namestr

def shortnamesfilter(obj, objvalue, context):
    (fnames, mnames, lnames, snames) = _names(objvalue, short=True)
    namestr = ''
    for n in fnames:
        namestr = _punctuate(namestr) + n
    for n in mnames:
        namestr = _punctuate(namestr) + n
    for n in lnames:
        namestr = _punctuate(namestr) + n
    if len(snames) > 0:
        namestr = _punctuate(namestr, ',')
    for n in snames:
        namestr = _punctuate(namestr) + n
    return namestr

def shortnameslistfilter(obj, objvalue, context):
    for i in range(len(objvalue)):
        objvalue[i] = shortnamesfilter(obj, objvalue[i], context)
    return objvalue

def lastfirstlistfilter(obj, objvalue, context):
    if objvalue:
        objvalue = copy(objvalue)
        objvalue[0] = lastfirstfilter(obj, objvalue[0], context)
    return objvalue

def alllastfirstlistfilter(obj, objvalue, context):
    if objvalue:
        objvalue = copy(objvalue)
        for i in range(len(objvalue)):
            objvalue[i] = lastfirstfilter(obj, objvalue[i], context)
    return objvalue

def alllastlistfilter(obj, objvalue, context):
    if objvalue:
        objvalue = copy(objvalue)
        for i in range(len(objvalue)):
            (fnames, mnames, lnames, snames) = _names(objvalue[i])
            objvalue[i] = ''
            for n in lnames:
                objvalue[i] = _punctuate(objvalue[i]) + n
    return objvalue

def plainlistformatter(obj, objvalue, context):
    value = ''
    for i in range(len(objvalue)):
        value = _punctuate(value, ',') + str(objvalue[i])
    return value

def commalistformatter(obj, objvalue, context):
    value = ''
    for i in range(len(objvalue)):
        if value:
            if len(objvalue) > 2:
                value += ','
            value += ' '
            if i == len(objvalue) - 1:
                value += 'and '
        value += str(objvalue[i])
    return value

def andlistformatter(obj, objvalue, context):
    return ' and '.join([str(element) for element in objvalue])

def andcrosstexlistformatter(obj, objvalue, context):
    return ' and '.join([isinstance(element, Object) and element._primarykey or str(element) for element in objvalue])

def initialslistformatter(obj, objvalue, context):
    value = ''
    if len(objvalue) == 1:
        value = _last_initials(objvalue[0], 3)
    elif len(objvalue) <= 4:
        for i in range(0, min(len(objvalue), 5)):
            value += _last_initials(objvalue[i], 1)
    else:
        for i in range(0, 4):
            value += _last_initials(objvalue[i], 1)
        value += '{\etalchar{+}}'
    return value

def fullnameslistformatter(obj, objvalue, context):
    value = ''
    if len(objvalue) == 2:
        (fnames1, mnames1, lnames1, snames1) = _names(objvalue[0])
        (fnames2, mnames2, lnames2, snames2) = _names(objvalue[1])
        value = ' '.join(mnames1 + lnames1) + ' \& ' + ' '.join(mnames2 + lnames2)
    elif objvalue:
        (fnames1, mnames1, lnames1, snames1) = _names(objvalue[0])
        value = ' '.join(mnames1 + lnames1)
        if len(objvalue) > 2:
            value += ' et al.'
    return value

def makebracketfilter(left, right):
    def bracketfilter(obj, objvalue, context):
        if objvalue:
            return '%s%s%s' % (left, objvalue.strip(), right)
        return objvalue
    return bracketfilter

def makesuffixfilter(suffix):
    def suffixfilter(obj, objvalue, context):
        if objvalue:
            return '%s%s' % (objvalue.strip(), suffix)
        return objvalue
    return suffixfilter

def edfilter(obj, objvalue, context):
    if objvalue:
        if ' and ' in objvalue:
            objvalue = objvalue + ', eds.'
        else:
            objvalue = objvalue + ', ed.'
    return objvalue

def makeprefixfilter(prefix):
    def prefixfilter(obj, objvalue, context):
        if objvalue:
            return '%s%s' % (prefix, objvalue.strip())
        return objvalue
    return prefixfilter

def bibitemfilter(obj, objvalue, context):
    if objvalue:
        label = obj._format(*(context + ('label',)))
        if label:
            label = '[%s]' % label
        return '\\bibitem%s{%s}\n%s\n\n' % (label, obj.citation, objvalue.strip())
    return objvalue

def emptyfilter(obj, objvalue, context):
    return ''

def makeuniquefilter():
    used = []
    def uniquefilter(obj, objvalue, context):
        if objvalue != '':
            if objvalue in used:
                for char in list('abcdefghijklmnopqrstuvwxyz'):
                    if objvalue + char not in used:
                        objvalue += char
                        break
                else:
                    raise ValueError, 'too many citations with key %s' % objvalue
            used.append(objvalue)
        return objvalue
    return uniquefilter

def twodigitfilter(obj, objvalue, context):
    return objvalue[-2:]

infilter = makeprefixfilter('In ')
procfilter = makeprefixfilter('Proc. of ')
proceedingsfilter = makeprefixfilter('Proceedings of the ')
emphfilter = makebracketfilter('\\emph{', '}')
boldfilter = makebracketfilter('\\textbf{', '}')
scfilter = makebracketfilter('\\textsc{', '}')
bracesfilter = makebracketfilter('{', '}')
quotefilter = makebracketfilter("``", "''")

def conferencetrackfilter(obj, objvalue, context):
    value = obj._format(*(context + ('conference',)))
    value = _punctuate(value, ',') + objvalue
    return value

def killfilter(obj, objvalue, context):
    if context[-1] in obj._required:
        return objvalue
    else:
        return ''

def titlecasefilter(obj, objvalue, context):
    newtitle = ''
    dollars = 0
    dashlen = 0
    inmath = False
    inliteral = False
    incommand = False
    wordbreak = True
    sentencebreak = True
    for i, char in enumerate(objvalue):
        if char == '{':
            close = _protectre.search(objvalue[i+1:])
            inliteral = not incommand and (close is not None and close.group() == '}')
        if char == '}':
            inliteral = False

        if char == '\\':
            incommand = True
        elif char.isspace():
            incommand = False

        if char == '-':
            dashlen += 1
        else:
            dashlen = 0

        if char == '$':
            dollars += 1
        elif dollars > 0:
            inmath = not inmath
            dollars = 0

        if not (inliteral or inmath or incommand):
            if wordbreak:
                newtitle += char.upper()
            else:
                newtitle += char.lower()
        else:
            newtitle += char

        sentencebreak = (not inliteral and not inmath and not incommand and (char in '!?:.' or dashlen > 1)) or (sentencebreak and (char.isspace() or incommand or inmath or char == '{'))
        wordbreak = sentencebreak or (not inliteral and not inmath and not incommand and (char.isspace() or char in ',-')) or (wordbreak and (incommand or inmath or char == '{'))

        if not char.isalnum() and char not in '_\\':
            incommand = False
    return newtitle

def lowertitlecasefilter(obj, objvalue, context):
    newtitle = ''
    dollars = 0
    dashlen = 0
    inmath = False
    inliteral = False
    incommand = False
    sentencebreak = True
    for i, char in enumerate(objvalue):
        if char == '{':
            close = _protectre.search(objvalue[i+1:])
            inliteral = not incommand and (close is not None and close.group() == '}')
        if char == '}':
            inliteral = False

        if char == '\\':
            incommand = True
        elif char.isspace():
            incommand = False

        if char == '-':
            dashlen += 1
        else:
            dashlen = 0

        if char == '$':
            dollars += 1
        elif dollars > 0:
            inmath = not inmath
            dollars = 0

        if not (inliteral or inmath or incommand):
            if sentencebreak:
                newtitle += char.upper()
            else:
                newtitle += char.lower()
        else:
            newtitle += char

        sentencebreak = (not inliteral and not inmath and not incommand and (char in '!?:.' or dashlen > 1)) or (sentencebreak and (char.isspace() or incommand or inmath or char == '{'))

        if not char.isalnum() and char not in '_\\':
            incommand = False
    return newtitle

def uppercasefilter(obj, objvalue, context):
    newtitle = ''
    dollars = 0
    dashlen = 0
    inmath = False
    inliteral = False
    incommand = False
    for i, char in enumerate(objvalue):
        if char == '{':
            close = _protectre.search(objvalue[i+1:])
            inliteral = not incommand and (close is not None and close.group() == '}')
        if char == '}':
            inliteral = False

        if char == '\\':
            incommand = True
        elif char.isspace():
            incommand = False

        if char == '-':
            dashlen += 1
        else:
            dashlen = 0

        if char == '$':
            dollars += 1
        elif dollars > 0:
            inmath = not inmath
            dollars = 0

        if not (inliteral or inmath or incommand):
            newtitle += char.upper()
        else:
            newtitle += char

        if not char.isalnum() and char not in '_\\':
            incommand = False
    return newtitle

def maketitlephrasefilter(titlephrases):
    def titlephrasefilter(obj, objvalue, context):
        newtitle = ''
        ignoreuntil = 0
        dollars = 0
        dashlen = 0
        inmath = False
        inliteral = False
        incommand = False
        wordbreak = True
        sentencebreak = True
        for i, char in enumerate(objvalue):
            if char == '{':
                close = _protectre.search(objvalue[i+1:])
                inliteral = not incommand and (close is not None and close.group() == '}')
            if char == '}':
                inliteral = False

            if char == '\\':
                incommand = True
            elif char.isspace():
                incommand = False

            if char == '-':
                dashlen += 1
            else:
                dashlen = 0

            if char == '$':
                dollars += 1
            elif dollars > 0:
                inmath = not inmath
                dollars = 0

            if i >= ignoreuntil:
                if wordbreak and not (inliteral or inmath or incommand):
                    match = ''
                    for phrase in titlephrases:
                        if objvalue.lower().startswith(phrase.lower(), i) and len(phrase) > len(match) and (i + len(phrase) >= len(objvalue) - 1 or not objvalue[i + len(phrase)].isalnum()):
                            match = phrase
                    if len(match) > 0:
                        ignoreuntil = i + len(match)
                        newtitle += match
                    else:
                        newtitle += char
                else:
                    newtitle += char

            sentencebreak = (not inliteral and not inmath and not incommand and (char in '!?:.' or dashlen > 1)) or (sentencebreak and (char.isspace() or incommand or inmath or char == '{'))
            wordbreak = sentencebreak or (not inliteral and not inmath and not incommand and (char.isspace() or char in ',-')) or (wordbreak and (incommand or inmath or char == '{'))

            if not char.isalnum() and char not in '_\\':
                incommand = False
        return newtitle
    return titlephrasefilter

def makelowerphrasefilter(lowerphrases):
    def lowerphrasefilter(obj, objvalue, context):
        newtitle = ''
        ignoreuntil = 0
        dollars = 0
        dashlen = 0
        inmath = False
        inliteral = False
        incommand = False
        wordbreak = True
        sentencebreak = True
        for i, char in enumerate(objvalue):
            if char == '{':
                close = _protectre.search(objvalue[i+1:])
                inliteral = not incommand and (close is not None and close.group() == '}')
            if char == '}':
                inliteral = False

            if char == '\\':
                incommand = True
            elif char.isspace():
                incommand = False

            if char == '-':
                dashlen += 1
            else:
                dashlen = 0

            if char == '$':
                dollars += 1
            elif dollars > 0:
                inmath = not inmath
                dollars = 0

            if i >= ignoreuntil:
                if wordbreak and not (sentencebreak or inliteral or inmath or incommand):
                    match = ''
                    for phrase in lowerphrases:
                        if objvalue.lower().startswith(phrase.lower(), i) and len(phrase) > len(match) and (i + len(phrase) >= len(objvalue) - 1 or not objvalue[i + len(phrase)].isalnum()):
                            match = phrase.lower()
                    if len(match) > 0:
                        ignoreuntil = i + len(match)
                        newtitle += match
                    else:
                        newtitle += char
                else:
                    newtitle += char

            sentencebreak = (not inliteral and not inmath and not incommand and (char in '!?:.' or dashlen > 1)) or (sentencebreak and (char.isspace() or incommand or inmath or char == '{'))
            wordbreak = sentencebreak or (not inliteral and not inmath and not incommand and (char.isspace() or char in ',-')) or (wordbreak and (incommand or inmath or char == '{'))

            if not char.isalnum() and char not in '_\\':
                incommand = False
        return newtitle
    return lowerphrasefilter

def listproducer(obj, value, context):
    if isinstance(obj, list):
        return list(obj)
    else:
        return None

ObjectList._addproducer(listproducer, 'value')
Object._addlistfilter(alllastfirstlistfilter, 'sort', 'author')
Object._addfilter(titlecasefilter, 'sort', 'author')



'''
Handle CrossTeX's many command-line options.

This is somewhat complicated further by the fact that options can be
set in .aux files on a per-document, on-the-fly basis using \bibstyle.
It is therefore a desirable property to build up options iteratively.
'''

import sys
import optparse
import crosstex.objects


def heading(option, opt, value, parser, reversed=False):
    parser.values.heading.append((reversed, value))

def sorting(option, opt, value, parser, reversed=False):
    if parser.values.sort == None or value == 'none':
        parser.values.sort = []
    if value != 'none':
        parser.values.sort.append((reversed, value))

def message(opts, level, details):
    if level >= opts.check:
        if level > 0:
            opts.errors += 1
        else:
            opts.warnings += 1
        sys.stderr.write(str(details) + '\n')

def warning(opts, details):
    message(opts, 0, details)

def error(opts, details):
    message(opts, 1, details)


_strings = [ name for name in dir(crosstex.objects) \
             if (isinstance(getattr(crosstex.objects, name), type) and \
                 issubclass(getattr(crosstex.objects, name), \
                            crosstex.objects.string)) ]

_dumpable = ['file', 'titlephrase', 'titlesmall'] + _strings


class OptionParser(optparse.OptionParser):

    def __init__(self, version='Testing', xtxlib=[]):
        optparse.OptionParser.__init__(self,
          usage='usage: %prog [options] files',
          version='CrossTeX %s' % version,
          description='A modern, object-oriented bibliographic tool.')

        self.set_defaults(errors=0)
        self.set_defaults(warnings=0)

        self.set_defaults(check=1)
        self.add_option('--quiet',
          action='store_const',
          const=2,
          dest='check',
          help='Do not sanity check the input.')
        self.add_option('--strict',
          action='store_const',
          const=0,
          dest='check',
          help='Apply stricter checks and check all entries.')

        self.set_defaults(dump=[])
        self.add_option('--dump',
          metavar='TYPE',
          action='append',
          type='choice',
          choices=_dumpable,
          help='After parsing the bibliography, dump a list of all objects of the '
               'type specified, or, with "file", print a list of files processed.')

        dirs = xtxlib
        if '.' not in dirs:
            dirs += '.'
        self.set_defaults(dir=dirs)
        self.add_option('-d', '--dir',
          type='string',
          action='append',
          help='Add a directory in which to find data files, searched from last '
               'specified to first. Default: %default')

        self.set_defaults(cite=[])
        self.add_option('--cite',
          type='string',
          action='append',
          help='Cite a key exactly as with the \\cite LaTeX command.')

        self.set_defaults(cite_by='style')
        self.add_option('--cite-by',
          type='choice',
          choices=['number', 'initials', 'fullname', 'style'],
          help='With "number", use numeric labels such as [1]. With "initials", '
               'use labels based on author last-names such as [SBRD04b]. With '
               '"fullname", use labels based on author names such as '
               '[Sirer et al. 2004]. With "style", use the style default.')

        self.set_defaults(style=None)
        self.add_option('--style',
          metavar='STYLE',
          type='string',
          help='Use a standard style such as plain, unsrt, abbrv, full, or alpha. '
               'Options set by the style may be overidden by further command-line '
               'options.')

        self.set_defaults(sort=[(False, 'title'), (False, 'monthno'),
                                (False, 'year'), (False, 'author')])
        self.add_option('-s', '--sort',
          metavar='FIELD',
          type='string',
          action='callback',
          callback=sorting,
          help='Sort by specified field. Multiple sort orders are applied in the '
               'order specified, e.g. "-s year -s author" will cause elements to '
               'be grouped primarily by author and sub-grouped by year.')
        self.add_option('-S', '--reverse-sort',
          metavar='FIELD',
          type='string',
          action='callback',
          callback=sorting,
          callback_args=(True,),
          help='Exactly as --sort, but sort by descending field values rather '
               'than ascending.')
        self.add_option('--no-sort',
          action='store_const',
          const=[],
          dest='sort')

        self.set_defaults(heading=[])
        self.add_option('--heading',
          metavar='FIELD',
          type='string',
          action='callback',
          callback=heading,
          help='Divide entries and create headings in bibliography by the value '
               'of the given field.')
        self.add_option('--reverse-heading',
          metavar='FIELD',
          type='string',
          action='callback',
          callback=heading,
          callback_args=(True,),
          help='Divide entries and create headings in bibliography by the value '
               'of the given field.')

        self.set_defaults(short=[])
        self.add_option('--short',
          metavar='TYPE',
          type='choice',
          action='append',
          choices=_strings,
          help='Specify any string-like object, i.e. one with name and shortname '
               'fields. Whenever possible,the short name will be used,e.g. '
               'two-letter state codes for "state",conference acronyms such as '
               'NSDI for "conference",or initials such as E. G. Sirer for '
               '"author".')

        self.set_defaults(capitalize=[])
        self.add_option('--capitalize',
          metavar='TYPE',
          type='choice',
          choices=_strings,
          action='append',
          help='Specify any string-like object, i.e. one with name and shortname '
               'fields. Strings of the specified types will appear in ALL CAPS.')

        self.set_defaults(no_field=[])
        self.add_option('--no-field',
          metavar='FIELD',
          action='append',
          help='Specify a field name, and in any objects where that field is '
               'optional it will be unassigned no matter what appears in the '
               'database.  For example, to turn off page numbers, use '
               '"--no-field pages".')

        self.set_defaults(link=[])
        self.add_option('-l', '--link',
          metavar='FIELD',
          type='string',
          action='append',
          help='Add to the list of fields used to generate links. LaTeX documents '
               'should make use of links by including the hyperref package. When '
               'converting to HTML, this defaults to [Abstract, URL, PS, PDF, '
               'HTML, DVI, TEX, BIB, FTP, HTTP, and RTF].')
        self.add_option('--no-link',
          action='store_const',
          const=[],
          dest='link')

        self.set_defaults(titlecase='default')
        self.add_option('--titlecase',
          type='choice',
          choices=['lower', 'upper', 'as-is', 'default'],
          help='In the bibliography, force titles into lower-, upper-, or '
               'title-case. Default: Leave titles unchanged.')

        self.set_defaults(convert='bbl')
        self.add_option('--xtx2bib',
          action='store_const',
          const='bib',
          dest='convert',
          help='Convert the bibliography information to old-style BibTeX.')
        self.add_option('--xtx2html',
          action='store_const',
          const='html',
          dest='convert',
          help='Format the bibliography as HTML.')
        self.add_option('--bib2xtx',
          action='store_const',
          const='xtx',
          dest='convert',
          help='Format the bibliography as HTML.')

        self.add_option('--add-in',
          action='store_true',
          help='Add "In" for articles.')
        self.add_option('--add-proc',
          action='store_true',
          help='Add "Proc. of" to conference and workshop publications.')
        self.set_defaults(add_in=[])
        self.add_option('--add-proceedings',
          action='store_true',
          help='Add "Proceedings of the" to conference and workshop publications.')
        self.add_option('--abstract',
          action='store_true',
          help='In the bibliography, include paper abstracts if available.')
        self.add_option('--no-abstract',
          action='store_false',
          dest='abstract')
        self.add_option('--keywords',
          action='store_true',
          help='In the bibliography, include paper keywords if available.')
        self.add_option('--no-keywords',
          action='store_false',
          dest='keywords')
        self.add_option('--popups',
          action='store_true',
          help='If abstracts or keywords are to appear for an entry when '
               'generating HTML, instead hide these extra blocks and reveal them '
               'as popups when the mouse hovers over the entry.')
        self.add_option('--no-popups',
          action='store_false')
        self.add_option('--title-head',
          action='store_true',
          help='In the bibliography, put the title bold and first.')
        self.add_option('--no-title-head',
          action='store_false',
          dest='title_head')
        self.add_option('--blank-labels',
          action='store_true',
          help='In the bibliography, leave out item labels.')
        self.add_option('--no-blank-labels',
          action='store_false',
          dest='blank_labels')
        self.add_option('--break-lines',
          action='store_true',
          help='In the bibliography, put author, title, and publication '
               'information on separate lines.')
        self.add_option('--no-break-lines',
          action='store_false',
          dest='break_lines')
        self.add_option('--last-first',
          action='store_true',
          help='The first name in each author list will appear "Last, First" '
               'instead of "First Last" (the latter is the default).')
        self.add_option('--no-last-first',
          action='store_false',
          dest='last_first')



from crosstex.style.common import *

# Standard producers for top-level objects.
publication._addproducer(makejoinproducer(".", "\n\\newblock ", ".", "", 'fullauthor', 'fulltitlepublicationextras'), 'value')
location._addproducer(citystatecountryproducer, 'value')
ObjectList._addlistformatter(commalistformatter, 'value')

# Long labels.
publication._addproducer(longauthoryearproducer, 'fullnamelabel') # Default to author and year, separated by space.
publication._addfilter(makeuniquefilter(), 'fullnamelabel') # Ensure unique labels by appending letters.
publication._addproducer(makegetterproducer('editor'), 'fullnamelabel', 'author') # Fall back to editor if there are no authors.
publication._addlistformatter(fullnameslistformatter, 'fullnamelabel', 'author') # Format authors as last names.

# Alphabetic labels.
publication._addproducer(authoryearproducer, 'initialslabel') # Default to author and year with no space between.
publication._addfilter(makeuniquefilter(), 'initialslabel') # Ensure unique labels by appending letters.
publication._addproducer(makegetterproducer('editor'), 'initialslabel', 'author') # Fall back to editor if there are no authors.
publication._addlistformatter(initialslistformatter, 'initialslabel', 'author') # Format authors as a concatenation of last initials.
publication._addfilter(twodigitfilter, 'initialslabel', 'year') # Format year with only two least-significant digits.

# By default, use key as label if provided.
publication._addproducer(makegetterproducer('key'), 'label')

# The virtual fullauthor field contains the list of authors (or editors, with no authors).
publication._addproducer(authoreditorproducer, 'fullauthor')

# The virtual fulltitle field contains the title.
publication._addproducer(makegetterproducer('title'), 'fulltitle')

# The virtual fullpublication field collects the rest of the relevant information.
publication._addproducer(fullpublicationproducer, 'fullpublication')
publication._addfilter(edfilter, 'fullpublication', 'editor') # Editors are denoted with 'ed.'

# The virtual fullextras field collects the miscellanous extra information.
publication._addproducer(makejoinproducer(".", " ", ".", "", 'links', 'extras'), 'fullextras')

# The following are various combinations of useful fields.
publication._addproducer(makejoinproducer(".", "\n\\newblock ", "", "", 'fullauthor', 'fulltitle'), 'fullauthortitle')
publication._addproducer(makejoinproducer(".", "\n\\newblock ", "", "", 'fulltitle', 'fullpublication'), 'fulltitlepublication')
publication._addproducer(makejoinproducer(".", "\n\\newblock ", "", "", 'fullpublication', 'fullextras'), 'fullpublicationextras')
publication._addproducer(makejoinproducer(".", "\n\\newblock ", "", "", 'fulltitle', 'fullpublication', 'fullextras'), 'fulltitlepublicationextras')
publication._addproducer(makejoinproducer(".", "\n\\newblock ", "", "", 'fullauthor', 'fullpublication', 'fullextras'), 'fullauthorpublicationextras')
publication._addproducer(makejoinproducer(".", "\n\\newblock ", "", "", 'fullauthor', 'fulltitle', 'fullpublication'), 'fullauthortitlepublication')

# Preface conference tracks and workshops with the name of the conference.
conferencetrack._addfilter(conferencetrackfilter, 'value')

# Specialized publication information.
publication._addproducer(makegetterproducer('howpublished'), 'publication')
url._addproducer(accessedproducer, 'publication')
thesis._addproducer(thesistypeproducer, 'publication')

# Specialized type information.
patent._addproducer('United States Patent', 'numbertype')
techreport._addproducer('Technical Report', 'numbertype')
rfc._addproducer('IETF Request for Comments', 'numbertype')
mastersthesis._addproducer("Master's", 'thesistype')
phdthesis._addproducer('Ph.D.', 'thesistype')

# Wrap publications in \bibitem entries.
publication._addfilter(bibitemfilter, 'value')

# The note field appears at the end of misc objects.
misc._addproducer(makejoinproducer(".", " ", ".", "", 'note', 'links', 'extras'), 'fullextras')
misc._addfilter(killfilter, 'fullpublication', 'note')



from crosstex.style.common import *

# Prefer long names.
string._addproducer(makegetterproducer('shortname'), 'value')
string._addproducer(makegetterproducer('longname'), 'value')

# Standard producers for top-level objects.
publication._addproducer(bibtexproducer, 'value')
location._addproducer(citystatecountryproducer, 'value')
ObjectList._addlistformatter(andlistformatter, 'value')

# Use empty labels (irrelevant, but labels determine what is 'citeable' and thus produced.
publication._addproducer(emptyproducer, 'label')

# Preface conference tracks and workshops with the name of the conference.
conferencetrack._addfilter(conferencetrackfilter, 'value')

# Specialized publication information.
publication._addproducer(makegetterproducer('howpublished'), 'publication')
url._addproducer(accessedproducer, 'publication')
thesis._addproducer(thesistypeproducer, 'publication')

# Specialized type information.
patent._addproducer('United States Patent', 'numbertype')
techreport._addproducer('Technical Report', 'numbertype')
rfc._addproducer('IETF Request for Comments', 'numbertype')
mastersthesis._addproducer("Master's", 'thesistype')
phdthesis._addfilter('Ph.D.', 'thesistype')



from crosstex.style.basic import *

# Use ACM publication syntax.
publication._addproducer(acmfullpublicationproducer, 'fullpublication')

# Prefer long names.
string._addproducer(makegetterproducer('shortname'), 'value')
string._addproducer(makegetterproducer('longname'), 'value')

# Use 'In' and emphasize book and journal titles.
publication._addfilter(emphfilter, 'fullpublication', 'booktitle')
publication._addfilter(emphfilter, 'fullpublication', 'journal')
publication._addfilter(infilter, 'fullpublication', 'booktitle')

# Use first initials and small caps for authors.
publication._addlistfilter(shortnameslistfilter, 'fullauthor', 'author')
publication._addlistfilter(alllastfirstlistfilter, 'fullauthor', 'author')
publication._addfilter(scfilter, 'fullauthor', 'author')
publication._addlistfilter(shortnameslistfilter, 'fullauthor', 'editor')
publication._addlistfilter(alllastfirstlistfilter, 'fullauthor', 'editor')
publication._addfilter(scfilter, 'fullauthor', 'editor')

# Sort on last names only.
publication._addlistfilter(alllastlistfilter, 'sort', 'author')
publication._addlistfilter(alllastlistfilter, 'sort', 'editor')
publication._addlistformatter(plainlistformatter, 'sort', 'author')
publication._addlistformatter(plainlistformatter, 'sort', 'editor')

# Use numeric labels.
publication._addproducer(emptyproducer, 'label')



from distutils.core import setup


classifiers = [ 'Development Status :: 5 - Production/Stable'
              , 'Intended Audience :: Science/Research'
              , 'License :: OSI Approved :: GNU General Public License v2 (GPLv2)'
              , 'Operating System :: MacOS :: MacOS X'
              , 'Operating System :: POSIX :: Linux'
              , 'Operating System :: Unix'
              ]

setup(name='CrossTex',
      version='0.7.0',
      author='Robert Escriva (rescrv)',
      author_email='escriva@cs.cornell.edu',
      packages=['crosstex'
               ,'crosstex.style'
               ],
      scripts=['bin/crosstex'],
      url='http://firmant.org/',
      license='GPLv2',
      description='CrossTeX is a bibliography management tool',
      classifiers=classifiers,
      )
